module Backend where

import Data.List
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid

import AST
import Dot
import Engine
import Instr
import Label
import Type

lns :: [String] -> String
lns = intercalate "\n"

scope :: String -> [String] -> String
scope s ss = lns [sss ++ "{", lns $ map ("  " ++) (concatMap lines ss), "}"]
  where
    sss | s == ""   = ""
        | otherwise = s ++ " "

data ResolvedLabel
  = Current Label
  | Other Int Label
  | Nowhere

compileEngine :: Map Label Node -> Map Node Int -> (Node, Engine) -> String
compileEngine labelMappings
              nodeMappings
              (currentNode, Engine name _ labels _) = lns
  [ "/* " ++ name ++ " */"
  , "#include \"prelude.h\""
  , "static inline Registers* code(Registers* regs, int label)" `scope`
      ["switch (label)" `scope` (map compileLabel $ M.toList labels)
      ,"return regs;"
      ]
  , "int main(int argc, char** argv)" `scope`
    [ "initialise(argc, argv);"
    , "while(1)" `scope`
      [ "Registers* regs;"
      , "int label = receive(&regs);"
      , "spawn_thread(regs, label);"
      ]
    ]
  ,""
  ]
  where

    compileLabel :: (Label, Code) -> String
    compileLabel (l@(Label _ n), code) = lns
      ["case " ++ show n ++ ":" `scope`
        ((show l ++ ":") : "PRINT_DEBUG_INFO;" : compileCode code ++ ["break;"])
      ]

    resolveLabel :: Label -> ResolvedLabel
    resolveLabel l = case node of
      Just n | n == currentNode -> Current l
             | otherwise        -> Other nodeNum l
      Nothing                   -> Nowhere
      where node    = M.lookup l labelMappings
            nodeNum = fromMaybe (error $ "resolveLabel: " ++ show l) $ flip M.lookup nodeMappings =<< node

    sparkend :: Label -> [String]
    sparkend l = case resolveLabel l of
      Current label          -> ["goto " ++ show label ++ ";"]
      Other node (Label _ n) -> ["send(regs, " ++ show node ++ ", " ++ show n ++ ");"]
      Nowhere                -> []

    spark :: Label -> [String]
    spark l = case resolveLabel l of
      Current (Label _ n)    -> ["fork_thread(regs, " ++ show n ++ ");"]
      Other node (Label _ n) -> ["send(regs, " ++ show node ++ ", " ++ show n ++ ");"]
      Nowhere                -> []

    ifzero :: Reg -> Code -> Code -> [String]
    ifzero r c1 c2 =
      ["if(" ++ reg r ++ "== 0)" `scope` compileCode c1
      ,"else" `scope` compileCode c2
      ]

    compileCode :: Code -> [String]
    compileCode code = case code of
      Spark l :. End -> {-["/* " ++ show (Spark l) ++ " */"] ++ -} sparkend l
      i :. rest      -> {-["/* " ++ show i ++ " */"] ++ -} compileInstr i ++ compileCode rest
      IfZero r c1 c2 -> ifzero r c1 c2
      Unpar r c      -> ["" `scope`
                          ["int_t tmp = " ++ "atomic_incr(" ++ reg r ++ ".mutex, 2);"
                          ,"if(tmp >= 2)" `scope` compileCode c
                          ]
                        ]
      End            -> []

    compileInstr :: Instr -> [String]
    compileInstr instr =  case instr of
      New r1 r2 l d    -> [reg d ++ ".ptr = new_heapval("
                                 ++ reg r1 ++ ", " ++ reg (Reference r2)
                                 ++ ", " ++ reg l ++ ");"
                          ]
      Get r l (d1, d2) -> ["get_heapval(" ++ reg r ++ ".ptr , " ++ reg (Reference d1)
                                  ++ ", " ++ reg (Reference d2) ++ ", " ++ reg l ++ ");"
                          ]
      Update d r l     -> [reg d ++ ".ptr = update_heapval(" ++ reg d ++
                                                         ".ptr, " ++ reg r ++
                                                         ", " ++ reg l ++
                                                         ");"
                          ]
      Free r           -> ["FREE(" ++ reg r ++ ".ptr);"
                          , reg r ++ ".integer = 0;"
                          ]
      Flip r1 r2       -> ["" `scope`
                            [ "Data tmp = " ++ reg r1 ++ ";"
                            , reg r1 ++ " = " ++ reg r2 ++ ";"
                            , reg r2 ++ " = tmp;"
                            ]
                          ]
      Set (LInt n) d    -> [reg d ++ ".integer = " ++ show n ++ ";"]
      Set (LFloat n1 n2) d -> [reg d ++ ".floating = " ++ show n1 ++ "." ++ show n2 ++ ";"]
      Prim (PrimOp p) r1 r2 d -> [p r1 r2 d ++ ";"]
      Par d            -> [reg d ++ ".mutex = malloc_mutex();"]
      Resize n         -> ["regs = resize_registers(regs, " ++ reg n ++ ");"]
      Spark l          -> spark l

resolvePorts :: Ord a => [(a, Engine)] -> Map Label a
resolvePorts es = M.fromList [(l, k) | (k, Engine _ _ ls _) <- es, l <- M.keys ls]

compileNet :: Net -> String -> ([(Node, String)], String)
compileNet sys@(Net es (TConstr (q@(Label _ initialPort), a@(Label _ _)) _)) rootCode =
  ( rootEngine : zip nodes (map (compileEngine labelMappings nodeMappings) nes)
  , dotEngines labelMappings nodeMappings (zip [0..] (Engine "root" Nothing undefined (TConstr (a, q) []) : es)))
  where
    rootNode = "root"
    nodes = zipWith (\n e -> fromMaybe n (engineLocus e))
                    (map (("n" ++) . show) [(1 :: Int) ..]) es
    nes   = zip nodes es
    labelMappings = M.fromList [(a, rootNode)] `mappend` resolvePorts nes
    nodeMappings  = M.fromList $ zip (rootNode : map fst nes) [0..]
    initialNode     = M.findWithDefault err q           labelMappings
    initialNodeRank = M.findWithDefault err initialNode nodeMappings
    err = error $ "compileNet: " ++ show sys
    rootEngine = (rootNode,
                  unlines ["#define INITIAL_NODE_RANK " ++ show initialNodeRank
                          ,"#define INITIAL_PORT "      ++ show initialPort
                          ] ++ rootCode)
compileNet s _ = error $ "compileNet: " ++ show s

createMakefile :: String -> String -> [String] -> String
createMakefile template flags files = lns $
  [ "FLAGS=" ++ flags
  , "SRC_FILES=" ++ unwords files
  , template
  ]
