{-# LANGUAGE TupleSections, TypeOperators #-}
module Compiler where

import Control.Applicative((<$>))
import Control.Monad
import Data.Function
import Data.List
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid
import Data.Ord

import AST
import Engine
import Interface
import Label
import Opt
import Pretty
import Type

infixr 5 :->:

type Context = Map Var Interface
data a :->: b = a :->: b deriving (Eq, Ord, Show)

--------------------------------------------------------------------------------
-- Combinators for working with nets "in a context"
--------------------------------------------------------------------------------
emptyCxt :: Net -> Context :->: Net
emptyCxt n = mempty :->: n

singletonCxt :: Var -> Net -> Context :->: Net
singletonCxt v (Net es (a1 :=>: a2)) =
  M.singleton v a1 :->: Net es a2
singletonCxt v n = error $ "singletonCxt: " ++ show (v, n)

cxtInterface :: Context -> Interface
cxtInterface = foldr (:*:) iunit . M.elems

interfaceCxt :: [Var] -> Interface -> Context
interfaceCxt [] (TVar ())      = mempty
interfaceCxt (v:vs) (i :*: is) = M.insert v i $ interfaceCxt vs is
interfaceCxt vs i              = error $ "interfaceCxt: " ++ show (vs, i)

cxtLambdaInv :: (Context :->: Net) -> (Net, [Var])
cxtLambdaInv (cxt :->: Net es a) =
  (Net es (cxtInterface cxt :=>: a), M.keys cxt)

cxtLambda :: [Var] -> Net -> Context :->: Net
cxtLambda vars (Net es (a :=>: b)) = interfaceCxt vars a :->: Net es b
cxtLambda vs n = error $ "cxtLambda: " ++ show (vs, n)

cxtProd :: Context :->: Net -> Context :->: Net -> LabelGen (Context :->: Net)
cxtProd (cxt1 :->: n1) (cxt2 :->: n2) = do
  let (n1', vars) = cxtLambdaInv (cxt1 `mappend` cxt2 :->: n1)
      (n2', _   ) = cxtLambdaInv (cxt2 `mappend` cxt1 :->: n2)
  cxtLambda vars <$> prod n1' n2'

cxtApp :: Context :->: Net -> Context :->: Net -> LabelGen (Context :->: Net)
cxtApp (cxt1 :->: n1) (cxt2 :->: n2) = do
  let dg1 = M.intersection cxt1 cxt2 -- The variables in the context that need to be
      dg2 = M.intersection cxt2 cxt1 -- contracted
      cxt1' = cxt1 M.\\ dg1 -- The variables that don't need it
      cxt2' = cxt2 M.\\ dg2
      dgi1   = cxtInterface dg1
      dgi2   = cxtInterface dg2
      dgvars = M.keys dg1
      (n1'@(Net _ i1), vars1) = cxtLambdaInv (cxt1' :->: n1)
      (n2'@(Net _ i2), vars2) = cxtLambdaInv (cxt2' :->: n2)
  app <- singleton <$> appEngine dgi1 i1 i2
  let Net es (dgres :*: g1res :*: g2res :=>: y) = lambdaInv $ pre lunitor
        ((lambda $ pre (dgi1 :*:) n1') `tensor`
         (lambda $ pre (dgi2 :*:) n2')) `compose` app
  return $ (interfaceCxt dgvars dgres `mappend`
            interfaceCxt vars1 g1res `mappend`
            interfaceCxt vars2 g2res) :->: Net es y

abstract :: Var -> Type -> Context :->: Net -> LabelGen (Context :->: Net)
abstract v t (cxt :->: Net es b) = do
  a <- case M.lookup v cxt of
    Just i  -> return i
    Nothing -> newInterface $ forgetVars t
  return $ M.delete v cxt :->: Net es (a :=>: b)

--------------------------------------------------------------------------------
-- Main compiler
--------------------------------------------------------------------------------
compile :: TypedTerm ::: Type -> LabelGen (Context :->: Net)
compile (TypedTerm term ::: typ) = case term of
  Var v -> do
    net <- singleton . modifyName ("id(" ++ v ++ ")")
        <$> idEngine (forgetVars typ)
    return $ singletonCxt v net

  Abs v (Just vtyp) t -> abstract v vtyp =<< compile t
  Abs _ Nothing _ -> err

  -- Some optimisations to do with pairs
  (TypedTerm ((TypedTerm Pair ::: _) :$ t1) ::: _) :$ t2 -> do
    cn1@(cxt1 :->: n1) <- compile t1
    cn2@(cxt2 :->: n2) <- compile t2
    case (M.null cxt1, M.null cxt2) of
      (True, _) -> return $ cxt2 :->: (n1 `tensorif` n2)
      (_, True) -> return $ cxt1 :->: (n1 `tensorif` n2)
      _         -> cxtProd cn1 cn2

  (TypedTerm Proj1 ::: _) :$ t -> do
    (cxt :->: Net n (typa :*: _typb)) <- compile t
    return $ cxt :->: Net n typa

  (TypedTerm Proj2 ::: _) :$ t -> do
    (cxt :->: Net n (_typa :*: typb)) <- compile t
    return $ cxt :->: Net n typb

  t1 :$ t2 -> do
    cn1@(cxt1 :->: n1@(Net _ nt1)) <- compile t1
    case nt1 of
      _ :=>: _ -> do
        cn2@(cxt2 :->: n2)                    <- compile t2
        -- if M.null cxt1 && M.null cxt2 then
        if M.null cxt2 then
          return $ cxt1 :->: lambdaInv (lambda n2 `compose` n1)
         else
         {-
          (p, vars) <- cxtLambdaInv <$> cxtProd cn2 cn1
          cxtLambda vars <$> gapp p
        -}
        {-
          (p, vars) <- cxtLambdaInv <$> cxtProd cn1 cn2
          ev <- netUncurry . singleton <$> idEngine (typa :=>: typb)
          cxtLambda vars <$> gcompose p ev
        -}
         cxtApp cn1 cn2
      _ -> error $ "application of non-function of type: " ++ show nt1 ++
                   " when compiling " ++ pretty (t1 :$ t2)

  Lit n -> emptyCxt . singleton <$> litEngine n

  Const c -> case cImpl c of
    Left (e, _, _) -> emptyCxt . singleton <$> e typ
    Right _ -> err

  Pair -> do
    let a :=>: b :=>: _ :*: _ = typ
    emptyCxt . netCurry . singleton . modifyName ","
      <$> idEngine (forgetVars (a :*: b))

  Proj1 -> do
    let a :*: b :=>: _ = forgetVars typ
    cc <- singleton . modifyName "proj1" <$> idEngine a
    sink <- bang b
    return $ emptyCxt $ post runitor (cc `tensor` sink)

  Proj2 -> do
    let a :*: b :=>: _ = forgetVars typ
    cc   <- singleton . modifyName "proj2" <$> idEngine b
    sink <- bang a
    return $ emptyCxt $ post lunitor (sink `tensor` cc)

  Fix -> do
    let ((a :=>: _) :=>: _) = forgetVars typ
    e <- modifyName "fix" <$> diagonalEngine a
    let (a1 :=>: a2 :*: a3) = engineInterface e
    return (emptyCxt $ Net [e] $ (a2 :=>: a1) :=>: a3)

  t :@ n -> withLocus (Just n) $ compile t
  where errs = "compile: " ++ show term ++ " : " ++ show typ
        err  = error errs

-- | Combine engines that are on the same locus
combine :: Net -> Net
combine (Net es i) = Net es' i
  where
    es' = concat [aux ses | ses <- groupBy ((==) `on` engineLocus)
                                  $ sortBy (comparing engineLocus) es]
    aux [] = error $ "combine: " ++ show es
    aux ses@(e:_) | isNothing (engineLocus e) = ses
                  | otherwise                 = [foldr1 engineUnion ses]

-- | Run the compiler and combine the results
evalCompiler :: TypedTerm ::: Type -> Net
evalCompiler term = if M.null cxt
  then combine net
  else error $ "evalCompiler: Not a closed term " ++ show term
  where cxt :->: net = evalLabelGen $ compile =<< optimise term
