module Engine where

import Control.Applicative
import Control.Exception
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid
import qualified Data.Set as S

import Label
import Type
import Instr
import Interface

--------------------------------------------------------------------------------
-- Engines and nets
--------------------------------------------------------------------------------
data Engine = Engine
  { engineName      :: String
  , engineLocus     :: Locus
  , engineCode      :: Map Label Code
  , engineInterface :: Interface
  }
  deriving (Eq, Ord)

modifyName :: String -> Engine -> Engine
modifyName n e = e {engineName = n}

engineUnion :: Engine -> Engine -> Engine
engineUnion (Engine name1 loc1 ls1 if1) (Engine name2 loc2 ls2 if2)
  | M.null $ M.intersection ls1 ls2 =
      Engine (name1 ++ "," ++ name2) loc1 (ls1 `mappend` ls2) (if1 :*: if2)
  | otherwise = error $ "engineUnion: " ++ show (ls1, ls2)

instance Show Engine where
  show (Engine name loc ls i) = unlines $
    ["Engine " ++ name ++ "@" ++ show loc] ++
    ["  " ++ show l ++ " -> " ++ show c | (l, c) <- M.toList ls] ++
    [show i]

instance LabelSubst Engine where
  labelSubst s (Engine n loc ls i) = Engine n loc (labelSubst s ls) (labelSubst s i)

data Net = Net [Engine] Interface
  deriving (Eq, Ord, Show)

engine :: String -> [(Label, Code)] -> Interface -> LabelGen Engine
engine name code i = do
  loc <- getLocus
  return $ Engine name loc (M.fromList code) i

singleton :: Engine -> Net
singleton e = Net [e] (engineInterface e)

bang :: TypeLike () bt -> LabelGen Net
bang t = do
  i <- newInterface t
  return (Net mempty (i :=>: iunit))

instance LabelSubst Net where
  labelSubst s (Net es i) = Net (labelSubst s es) (labelSubst s i)

infixr 0 -->
(-->) :: a -> b -> (a, b)
a --> b = (a, b)

--------------------------------------------------------------------------------
-- Game engines
--------------------------------------------------------------------------------
-- | Generalised copycat
copycatGen :: (Code -> Code) -> Interface -> Interface -> [(Label, Code)]
copycatGen init2 if1 if2 =
  let ini2 = initial if2
      opp = opponent (if1 :=>: if2)
      u21 = unifyInterface if1 if2
      u12 = unifyInterface if2 if1
      uni u r = M.findWithDefault (error $ "copycati: " ++ show (if1, if2)) r u in
   [r2 --> init2 $ sparkE r1
   | r2 <- S.toList ini2, let r1 = uni u21 r2] ++
   [r2 --> ccq $ sparkE r1
   | r2 <- S.toList (S.intersection opp (question if2) S.\\ ini2), let r1 = uni u21 r2] ++
   [r2 --> cca $ sparkE r1
   | r2 <- S.toList (S.intersection opp $ answer if2), let r1 = uni u21 r2] ++
   [r1 --> ccq $ sparkE r2
   | r1 <- S.toList (S.intersection opp (question if1)), let r2 = uni u12 r1] ++
   [r1 --> cca $ sparkE r2
   | r1 <- S.toList (S.intersection opp $ answer if1), let r2 = uni u12 r1]

copycati :: Interface -> Interface -> [(Label, Code)]
copycati = copycatGen cci

-- | Copycat engine
idEngine :: Show bt => TypeLike () bt -> LabelGen Engine
idEngine typ = do
  i@(a1 :=>: a2) <- newInterface $ typ :=>: typ
  let ls = copycati a1 a2
  engine "id" ls i

composeEngine :: Show bt => TypeLike () bt -> TypeLike () bt -> LabelGen Engine
composeEngine (x :=>: y1) (y2 :=>: z) = do
  () <- assert (shape y1 == shape y2) $ return ()
  i@((a1 :=>: b1) :*: (b2 :=>: c1) :=>: (a2 :=>: c2)) <- newInterface
    ((x  :=>: y1) :*: (y2 :=>: z)  :=>: (x  :=>: z))
  let ls = copycati c1 c2 ++ copycatGen exq a2 a1 ++ copycatGen exi b1 b2
  engine "compose" ls i
composeEngine x y = error $ "composeEngine: " ++ show (x, y)

diagonalGen :: (Code -> Code) -> (Code -> Code)
            -> Interface -> Interface -> Interface -> [(Label, Code)]
diagonalGen init2 init3 a1 a2 a3 =
  let i = a1 :=>: a2 :*: a3
      ini = initial i
      opp = opponent i
      u21 = unifyInterface a1 a2
      u31 = unifyInterface a1 a3
      u12 = unifyInterface a2 a1
      u13 = unifyInterface a3 a1
      uni u r = M.findWithDefault (error $ "diagonalGen: " ++ show i) r u
   in [r2 --> tempReg <~ Set (LInt 0) :. init2 (sparkE r1)
      | r2 <- S.toList ini, M.member r2 u21, let r1 = uni u21 r2] ++
      [r3 --> tempReg <~ Set (LInt 1) :. init3 (sparkE r1)
      | r3 <- S.toList ini, M.member r3 u31, let r1 = uni u31 r3] ++
      [r1 --> ccq $ IfZero (tempReg `Field` "integer") (sparkE r2) (sparkE r3)
      | r1 <- S.toList (S.intersection opp $ question a1)
      , let r2 = uni u12 r1, let r3 = uni u13 r1] ++
      [r1 --> cca $ IfZero (tempReg `Field` "integer") (sparkE r2) (sparkE r3)
      | r1 <- S.toList (S.intersection opp $ answer a1)
      , let r2 = uni u12 r1, let r3 = uni u13 r1] ++
      [r2 --> ccq $ sparkE r1
      | r2 <- S.toList (S.intersection opp (question a2) S.\\ ini), let r1 = uni u21 r2] ++
      [r3 --> ccq $ sparkE r1
      | r3 <- S.toList (S.intersection opp (question a3) S.\\ ini), let r1 = uni u31 r3] ++
      [r2 --> cca $ sparkE r1
      | r2 <- S.toList (S.intersection opp $ answer a2), let r1 = uni u21 r2] ++
      [r3 --> cca $ sparkE r1
      | r3 <- S.toList (S.intersection opp $ answer a3), let r1 = uni u31 r3]

diagonalEngine :: Show bt => TypeLike () bt -> LabelGen Engine
diagonalEngine typ = do
  i@(a1 :=>: a2 :*: a3) <- newInterface $ typ :=>: typ :*: typ
  engine "diagonal" (diagonalGen cci cci a1 a2 a3) i

appEngine :: Show bt => TypeLike () bt -> TypeLike () bt -> TypeLike () bt -> LabelGen Engine
appEngine dg (g :=>: x1 :=>: y) (g' :=>: x2) = do
  () <- assert (shape x1 == shape x2) $ return ()
  i@((dg1 :*: g1 :=>: a1 :=>: b1) :*: (dg2 :*: g2 :=>: a2) :=>: dg3 :*: g31 :*: g32 :=>: b3) <- newInterface
    ((dg  :*: g  :=>: x1 :=>: y)  :*: (dg  :*: g' :=>: x2) :=>: dg :*: g :*: g' :=>: y)
  let ls = copycati b1 b3 ++ diagonalGen ccqe exq dg3 dg1 dg2
         ++ copycatGen ccq g31 g1 ++ copycatGen exq g32 g2
         ++ copycatGen exi a2 a1
  engine "@" ls i

seqEngine :: LabelGen Engine
seqEngine = do
  i@(TConstr (r1, d1) _ :=>: TConstr (r2, d2) _ :=>: TConstr (r3, d3) _)
    <- newInterface $ tcom :=>: tcom :=>: tcom
  let ls =
        [ r3 --> cci $ sparkE r1
        , d1 --> cca $ Flip p1Reg p2Reg :. cci (sparkE r2)
        , d2 --> cca $ sparkE d3
        ]
  engine "seq" ls i

skipEngine :: LabelGen Engine
skipEngine = do
  i@(TConstr (r1, d1) _) <- newInterface tcom
  let ls =
        [ r1 --> Flip p1Reg p2Reg :. sparkE d1 ]
  engine "skip" ls i

litEngine :: Lit -> LabelGen Engine
litEngine lit = do
  let arr = zip [0..] $ litArrayView lit
      len = length arr
  i@(TConstr (r1, d1) _) <- newInterface tint
  let ls =
        [ r1 --> Flip p1Reg p2Reg :. p2Reg <~ Set (LInt 0) :. Resize (constReg len)
          :. foldr (\(n, l) rest -> dataReg `Index` show n <~ Set l :. rest) (sparkE d1) arr
        ]
  engine ("lit " ++ show lit) ls i

binOpEngine :: String -> PrimOp -> LabelGen Engine
binOpEngine opName op = do
  i@(TConstr (q1, a1) _ :=>: TConstr (q2, a2) _ :=>: TConstr (q3, a3) _) <- newInterface $
    tint :=>: tint :=>: tint
  let ls =
       [ q3 --> cci $ sparkE q1
       , a1 --> cca $ Flip p1Reg p2Reg :. cci (Update p2Reg dataReg dataLengthReg :. Resize const0 :. sparkE q2)
       , a2 --> (emptyReg, tempReg) <~ Get p1Reg const1 :. dataReg <~ Prim op tempReg dataReg :. cca (sparkE a3)
       ]
  engine opName ls i

unOpEngine :: String -> PrimOp -> LabelGen Engine
unOpEngine opName op = do
  i@(TConstr (q1, a1) _ :=>: TConstr (q2, a2) _) <- newInterface $
    tint :=>: tint
  let ls =
       [ q2 --> cci $ sparkE q1
       , a1 --> cca $ dataReg <~ Prim op dataReg emptyReg :. sparkE a2
       ]
  engine opName ls i

newvarEngine :: LabelGen Engine
newvarEngine = do
  i@((TConstr (q1, a1) _ :*: (TConstr (q2, a2) _ :=>: TConstr (q3, a3) _) :=>: TConstr (q4, a4) _)
    :=>: TConstr (q5, a5) _) <- newInterface ((tvar tint :=>: tint) :=>: tint)
  let ls =
       [ q5 --> tempReg <~ Set (LInt 0) :. cci (sparkE q4)
       , q1 --> Resize const1 :. (emptyReg, dataReg0) <~ Get p1Reg const1 :. Flip p1Reg p2Reg
             :. p2Reg <~ Set (LInt 0) :. sparkE a1
       , q3 --> Flip p1Reg p2Reg :. p2Reg <~ New p1Reg p2Reg const1 :. sparkE q2
       , a2 --> (emptyReg, tempReg) <~ Get p1Reg const1
                :.  Update tempReg dataReg dataLengthReg
                :. Resize const0 :. cca (sparkE a3)
       , a4 --> cca $ sparkE a5
       ]
  engine "newvar" ls i

mallocHeapVal = Prim $ PrimOp
  (\r1 r2 d -> reg d ++ ".ptr = malloc_heapval("
                    ++ reg r1 ++ ", " ++ reg r2 ++ ".integer)")
updateHeapArrayIndex = Prim $ PrimOp
  (\r1 r2 r3 -> "update_heap_array_index(" ++ reg r1 ++ ".ptr, " ++ reg r2 ++ ", " ++ reg r3 ++ ")")

accessHeapArrayIndex = Prim $ PrimOp
  (\r1 r2 d -> reg d ++ " = access_heap_array_index("
                    ++ reg r1 ++ ".ptr, " ++ reg r2 ++ ".integer)")

newarrayEngine :: LabelGen Engine
newarrayEngine = do
  i@(TConstr (qsz, asz) _
   :=>: ((TConstr (qidx, aidx) _ :=>: TConstr (qacc, aacc) _
                              :*: (TConstr (qupdval, aupdval) _ :=>: TConstr (qupd, aupd) _))
      :=>: TConstr (qinnerd, ainnerd) _)
   :=>: TConstr (qd, ad) _) <- newInterface (tint :=>: (tmutarray tint :=>: tint) :=>: tint)
  let ls =
       [ qd  --> cci (sparkE qsz)
       , asz --> cca $ p2Reg <~ mallocHeapVal p1Reg dataReg0 :. Resize const0 :. sparkE qinnerd
       , qacc --> Flip p1Reg p2Reg :. p2Reg <~ New p1Reg p2Reg const1 :. sparkE qidx
       , aidx --> IfZero (Dereference (p1Reg `Field` "ptr") `Field` "dataLength - 1")
                    ( (emptyReg, tempReg) <~ Get p1Reg const1 :.
                      dataReg0 <~ accessHeapArrayIndex tempReg dataReg0 :.
                      cca (sparkE aacc)
                    )
                    ( updateHeapArrayIndex p1Reg const1 dataReg0 :.
                      Resize const0 :.
                      (p2Reg, emptyReg) <~ Get p1Reg const0 :.
                      Flip p1Reg p2Reg :. sparkE qupdval
                    )
       , qupd --> Flip p1Reg p2Reg :. p2Reg <~ New p1Reg p2Reg (constReg 2) :. sparkE qidx
       , aupdval --> Flip tempReg dataReg0 :. Resize (constReg 2)
                  :. (p2Reg, dataReg0) <~ Get p1Reg (constReg 2)
                  :. updateHeapArrayIndex dataReg0 (dataReg `Index` show 1 `Field` "integer") tempReg
                  :. Flip p1Reg p2Reg :. p2Reg <~ Set (LInt 0) :. Resize const0 :. sparkE aupd
       , ainnerd --> cca $ sparkE ad
       ]
  engine "newarray" ls i

ifEngine :: LabelGen Engine
ifEngine = do
  i@(TConstr (q1, a1) _ :=>: TConstr (q2, a2) _ :=>: TConstr (q3, a3) _ :=>: TConstr (q4, a4) _)
    <- newInterface (tint :=>: tint :=>: tint :=>: tint)
  let ls =
       [ q4 --> cci $ sparkE q1
       , a1 --> cca $ Flip p1Reg p2Reg :. cci (IfZero (dataReg0 `Field` "integer") (Resize const0 :. sparkE q3) (Resize const0 :. sparkE q2))
       , a2 --> cca $ sparkE a4
       , a3 --> cca $ sparkE a4
       ]
  engine "if" ls i

parEngine :: LabelGen Engine
parEngine = do
  i@(TConstr (q1, a1) _ :=>: TConstr (q2, a2) _ :=>: TConstr (q3, a3) _)
    <- newInterface (tcom :=>: tcom :=>: tcom)
  let ls =
       [ q3 --> tempReg <~ Par :. cci (Spark q1 :. sparkE q2)
       , a1 --> (emptyReg, tempReg) <~ Get p1Reg const1 :. Unpar tempReg (cca $ sparkE a3)
       , a2 --> (emptyReg, tempReg) <~ Get p1Reg const1 :. Unpar tempReg (cca $ sparkE a3)
       ]
  engine "par" ls i

--------------------------------------------------------------------------------
-- Some useful combinators and identities
--------------------------------------------------------------------------------
-- | HRAM composition
compose :: Net -> Net -> Net
compose (Net es1 (a :=>: b)) (Net es2 (b' :=>: c)) =
  let sub = unifyInterface b b'
   in Net (es1 ++ labelSubst sub es2) (a :=>: c)

tensor :: Net -> Net -> Net
tensor (Net es1 (a1 :=>: b1)) (Net es2 (a2 :=>: b2)) =
  Net (es1 ++ es2) (a1 :*: a2 :=>: b1 :*: b2)

tensorif :: Net -> Net -> Net
tensorif (Net es1 a) (Net es2 b) =
  Net (es1 ++ es2) (a :*: b)

lambda :: Net -> Net
lambda (Net es a) = Net es (iunit :=>: a)

lambdaInv :: Net -> Net
lambdaInv (Net es (TVar () :=>: a)) = Net es a

netCurry :: Net -> Net
netCurry (Net es (a :*: b :=>: c)) = Net es (a :=>: b :=>: c)

netUncurry :: Net -> Net
netUncurry (Net es (a :=>: b :=>: c)) = Net es (a :*: b :=>: c)

pre :: (Interface -> Interface) -> Net -> Net
pre f (Net es (a :=>: b)) = Net es (f a :=>: b)

post :: (Interface -> Interface) -> Net -> Net
post f (Net es (a :=>: b)) = Net es (a :=>: f b)

lunitor :: Interface -> Interface
lunitor (TVar () :*: b) = b

lunitorinv :: Interface -> Interface
lunitorinv b = TVar () :*: b

runitor :: Interface -> Interface
runitor (a :*: TVar ()) = a

runitorinv :: Interface -> Interface
runitorinv a = a :*: TVar ()

-- | GAM composition
gcompose :: Net -> Net -> LabelGen Net
gcompose n1@(Net _ t1) n2@(Net _ t2) = do
  gcomp <- singleton <$> composeEngine t1 t2
  return $ lambdaInv $
    pre lunitor (lambda n1 `tensor` lambda n2) `compose` gcomp

{-
-- | GAM application
gapp :: Net -> LabelGen Net
gapp (Net es (gamma :=>: a :*: (a' :=>: b))) = do
  gcomp <- singleton <$> composeEngine (iunit :=>: a) (a' :=>: b)
  let Net es' (gamma' :=>: (TVar () :=>: b')) =
       Net es (gamma :=>: (iunit :=>: a) :*: (a' :=>: b)) `gcompose` gcomp
  return $ Net es' (gamma' :=>: b')
-}

-- | GAM identity
gid :: Show bt => TypeLike () bt -> LabelGen Net
gid t = singleton <$> idEngine t

gunit :: Show bt => TypeLike () bt -> LabelGen Net
gunit t = lambda <$> gid t

-- | <-,-> operator
prod :: Net -> Net -> LabelGen Net
prod n1@(Net _ (a1 :=>: _)) n2 = do
  diag <- singleton <$> diagonalEngine a1
  diag `gcompose` (n1 `tensor` n2)
