{-# LANGUAGE DeriveDataTypeable, DoRec #-}
module Parser where

import Data.Data

import Data.Parser.Grempa.Grammar

import Lexer

data OpList a
  = Base a
  | Another a String (OpList a)
  deriving (Eq, Show, Typeable)

data PType
  = PTOp (OpList PType)
  | PTCon String [PType]
  deriving (Eq, Show, Typeable)

ptcon :: String -> PType
ptcon = flip PTCon []

data Expr
  = EApp Expr Expr
  | EMacro String Expr Expr
  | EMapReduce String Integer Expr Expr
  | EVar String
  | EFix (String, Maybe PType) Expr
  | EInt Integer
  | EArray Expr
  | EFloat Integer Integer
  | ELam [(String, Maybe PType)] Expr
  | EAtNode Expr String
  | EOp (OpList Expr)
  deriving (Eq, Show, Typeable)

fun :: Grammar Tok Expr
fun = do
  rec
    -- Basic atoms
    var <- rule [fromTok <@> variable]
    con <- rule [fromTok <@> constructor]
    op  <- rule [fromTok <@> operation]
    num <- rule [fromNum <@> number]

    -- Types
    typeAtom <- rule [paren typ
                     ,ptcon <@> con
                     ,PTCon "[]" . (: []) <@ LBracket <#> typ <# RBracket
                     ]
    typeOpList <- rule [Another <@> typeAtom <#> op <#> typeOpList
                       ,Base <@> typeAtom
                       ]
    typ <- rule [PTOp <@> typeOpList]

    optionalAtNode <- rule [ epsilon id
                           , flip EAtNode <@ At <#> var
                           ]

    -- Vars
    typedVar   <- rule [typed <@> var <# Colon <#> typ]
    untypedVar <- rule [untyped <@> var]
    parenTypedVar <- rule [paren typedVar]
    maybeTypedVar <- rule [id <@> parenTypedVar
                          ,id <@> untypedVar
                          ]
    maybeTypedVars0 <- several0 maybeTypedVar
    maybeTypedVars  <- maybeTypedVar `cons` maybeTypedVars0
    varList <- rule [id <@> maybeTypedVars
                    ,(:[]) <@> typedVar
                    ]
    -- Expressions
    exprAtom <- levels $ do
      rec
        exprAtom0 <- lrule [ebang <@ Bang <#> exprAtom0]
        exprAtom1 <- lrule [EApp <@> exprAtom1 <#> exprAtom2
                           ,EMapReduce <@ MapReduce <# At <#> var
                            <#> num <#> exprAtom2 <#> exprAtom2]
        exprAtom2 <- lrule [EVar <@> var
                           ,EInt <@> num
                           ,EFloat <@> num <# Dot <#> num
                           ,EArray <@ LBracket <#> expr <# RBracket
                           ,EAtNode <@> exprAtom2 <# At <#> var
                           ,paren expr
                           ]
      return exprAtom0
    exprOpList <- rule [Another <@> exprAtom <#> op <#> exprOpList
                       ,Base <@> exprAtom
                       ]
    expr <- levels $ do
      rec
        expr0 <- lrule [elet <@ Let <#> maybeTypedVar <#> optionalAtNode <#> maybeTypedVars0 <# Equals
                              <#> expr0 <# In <#> expr0
                       ,emacro <@ Macro <#> var <#> maybeTypedVars0 <# Equals
                              <#> expr0 <# In <#> expr0
                       ,EFix <@ Fix    <#> maybeTypedVar <# Dot <#> expr0
                       ,ELam <@ Lambda <#> varList <# Dot <#> expr0
                       ]
        _expr2 <- lrule [eif <@ If <#> expr3 <# Then <#> expr0 <# Else <#> expr0]
        expr3  <- lrule [(\a o l -> EOp (Another a o l)) <@> exprAtom <#> op <#> exprOpList]
        _expr4 <- lrule [id <@> exprAtom]
      return expr0
  return expr
  where
    elet f atnode xs e e' = ELam [f] e' `EApp` (atnode $ ELam xs e)
    emacro f xs e e' = EMacro f (ELam xs e) e'
    paren x = id <@ LParen <#> x <# RParen
    typed var typ = (var, Just typ)
    untyped var   = (var, Nothing)
    eif a b c = ((EVar "if then else" `EApp` a) `EApp` b) `EApp` c
    ebang e = EVar "!" `EApp` e
