{-# LANGUAGE TypeOperators #-}
module TypeChecker where

import Control.Monad.State
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid

import AST
import Instr
import Pretty
import Type
import Unify

type TypeChecker a = State [Var] a
evalTypeChecker :: TypeChecker a -> a
evalTypeChecker = flip evalState $ map (("t" ++) . show) [(0 :: Int)..]

newTVar :: TypeChecker Type
newTVar = do
  v:vs <- get
  put vs
  return (TVar v)

type Cxt = Map Var Type

typeCheckOuter :: Term -> TypedTerm ::: Type
typeCheckOuter t = case mapMaybe tryType baseTypes of
    []    -> error $ "Term doesn't have ground type, it has type: " ++ pretty typ
    res:_ -> res
  where
    baseTypes = [tint, tfloat, tcom, tarray tint, tarray tfloat]
    (ttt, sub)  = evalTypeChecker $ typeCheck t mempty
    tt  ::: typ = typeSubst sub ttt
    tryType btyp
      | typ' == btyp = Just (tt' ::: typ')
      | otherwise    = Nothing
      where tt' ::: typ' = typeSubst (unify typ btyp) (tt ::: typ)

{-
desugarIfs :: TypedTerm ::: Type -> TypedTerm :: Type
desugarIfs ttyp@(t ::: typ) = case t of
  Var v                               -> ttyp
  Abs v mtype t                       -> (Abs v mtype $ desugarIfs t) ::: typ
  (((Const "if then else" ::: typ)
    :$ (a ::: typ1)) :$ (b ::: typ2)) :$ (c ::: typ3) -> case typ2 of
      typ21 :=>: typ22 ->
      typ
  t1 :$ t2                            -> (desugarIfs t1 :$ desugarIfs t2) ::: typ
  Lit l                               -> ttyp
  Const c                             -> ttyp
  Pair                                -> ttyp
  Proj1                               -> ttyp
  Proj2                               -> ttyp
  Fix                                 -> ttyp
  t :@ n                              -> (desugarIfs t :@ n) ::: typ
-}

typeCheck :: Term -> Cxt -> TypeChecker (TypedTerm ::: Type, TSubst)
typeCheck (Term term) cxt = case term of
  Var v -> do
    let typ = M.findWithDefault (error $ "Not in scope: " ++ v) v cxt
    return (Var v .:: typ, mempty)

  Abs v mtype1 t -> do
    type1 <- maybe newTVar return mtype1
    let cxt' = M.insert v type1 cxt
    (tt ::: type2, sub) <- typeCheck t cxt'
    return (Abs v (Just type1) (tt ::: type2) .:: type1 :=>: type2, sub)

  t1 :$ t2 -> do
    restype <- newTVar
    (tt1@(_ ::: funtype), subst1) <- typeCheck t1 cxt
    (tt2@(_ ::: argtype), subst2) <- typeCheck t2 (typeSubst subst1 cxt)
    let subst3  = uncurry unify $ typeSubst (mconcat [subst2, subst1])
                                    (funtype, argtype :=>: restype)
    let sub = mconcat [subst3, subst2, subst1]

    return (tt1 :$ tt2 .:: restype, sub)

  Lit (PrimLit (LInt n))     -> return (Lit (lint n) .:: tint, mempty)
  Lit (PrimLit (LFloat i j)) -> return (Lit (lfloat i j) .:: tfloat, mempty)
  Lit (LArray [])            -> do
    a <- newTVar
    return (Lit (LArray []) .:: TConstr TArray [a], mempty)
  Lit (LArray (l:ls))        -> do
    (TypedTerm (Lit (PrimLit l')) ::: a, _)
      <- typeCheck (Term $ Lit (PrimLit l)) cxt
    (TypedTerm (Lit (LArray ls')) ::: TConstr TArray [a'], _)
      <- typeCheck (Term $ Lit (LArray ls)) cxt
    let sub = unify a a'
    return (Lit (LArray (l':ls')) .:: TConstr TArray [typeSubst sub a'], sub)

  Const c -> case cImpl c of
    Left (_, neededVars, typFun) -> do
      tvars <- replicateM neededVars newTVar
      return (Const c .:: typFun tvars, mempty)
    Right _                      -> error $ "typeCheck.Const.Right: " ++ show term

  Pair -> do
    a <- newTVar
    b <- newTVar
    return (Pair .:: a :=>: b :=>: a :*: b, mempty)

  Proj1 -> do
    a <- newTVar
    b <- newTVar
    return (Proj1 .:: a :*: b :=>: a, mempty)

  Proj2 -> do
    a <- newTVar
    b <- newTVar
    return (Proj2 .:: a :*: b :=>: b, mempty)

  Fix -> do
    a <- newTVar
    return (Fix .:: (a :=>: a) :=>: a, mempty)

  t :@ n           -> do
    (tt@(_ ::: typ), sub) <- typeCheck t cxt
    return (tt :@ n .:: typ, sub)

  where
    infix 2 .::
    tt .:: typ = TypedTerm tt ::: typ
