{-# LANGUAGE TypeOperators #-}
module Compiler where
import Control.Applicative
import Control.Monad.State
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid
import Debug.Trace

import AST
import Component hiding (Instr(..))

data MsgType
  = MsgInt
  deriving (Eq, Ord, Show)

data CompilerState = CompilerState
  { csPorts           :: [Port]
  , csComponents      :: [(ComponentCode, Maybe Node)]
  , csConnectedPorts  :: [(Port, Port)]
  , csNode            :: Maybe Node
  }

initialCompilerState :: CompilerState
initialCompilerState = CompilerState
  { csPorts          = [0..]
  , csComponents     = mempty
  , csConnectedPorts = mempty
  , csNode           = mempty
  }

data Interface
  = In Interface Interface
  | Port MsgType Port
  deriving (Eq, Ord, Show)

ports :: Interface -> ([Port], [Port])
ports (In i o)   = (is1 `mappend` is2, os1 `mappend` os2)
  where (os1, is1) = ports i
        (is2, os2) = ports o
ports (Port t p) = (mempty, [p])

newInterface :: Type -> Compiler Interface
newInterface typ = case typ of
  TInt       -> do
    p <- newPort
    return $ Port MsgInt p
  t1 :->: t2 -> In <$> newInterface t1 <*> newInterface t2
  _          -> error $ "newInterface:" ++ show typ

reconstruct :: Interface -> ([Port], [Port]) -> Interface
reconstruct i ps = fst $ go i ps
  where
    go :: Interface -> ([Port], [Port]) -> (Interface, ([Port], [Port]))
    go (In i o) (is, os) =
      let (i', (is', os'))  = go i (os, is)
          (o', ps)          = go o (os', is')
       in (In i' o', ps)
    go (Port t p) (is, o:os) = (Port t o, (is, os))

type Compiler a = StateT CompilerState ComponentGen a

evalCompiler :: Compiler a -> (a, ([(ComponentCode, Maybe Node)], [(Port, Port)]))
evalCompiler c =
  let (a, x) = evalComponent $ runStateT c initialCompilerState
   in (a, (csComponents x, csConnectedPorts x))

withNode :: Node -> Compiler a -> Compiler a
withNode n m = do
  s <- get
  put s {csNode = Just n}
  r <- m
  modify $ \s' -> s' {csNode = csNode s}
  return r

newPort :: Compiler Port
newPort = do
  p:ps <- gets csPorts
  modify $ \s -> s {csPorts = ps}
  return p

emit :: ComponentCode -> Compiler ()
emit c = modify $ \s -> s {csComponents = (c, csNode s) : csComponents s}

connect :: Port -> Port -> Compiler ()
connect p1 p2 = modify $ \s -> s {csConnectedPorts = (p1, p2) : csConnectedPorts s}

connectInterface :: Interface -> Interface -> Compiler ()
connectInterface (In i1 o1) (In i2 o2) = do
  connectInterface i1 i2
  connectInterface o1 o2
connectInterface (Port t1 p1) (Port t2 p2) = do
  unless (t1 == t2) $ error $ "connectInterface: " ++ show (t1, t2)
  connect p1 p2
connectInterface t1 t2 = error $ "connectInterface: " ++ show (t1, t2)

apply :: Interface -> Interface -> Compiler Interface
apply (In i1 o1) i = do
  connectInterface i1 i
  return o1

addComponent :: ([Port] -> [Port] -> ComponentGen ComponentCode)
             -> Interface
             -> Compiler Interface
addComponent component i = head <$> addComponentIs component [i]

addComponentIs :: ([Port] -> [Port] -> ComponentGen ComponentCode)
               -> [Interface]
               -> Compiler [Interface]
addComponentIs component is = case is of
  [] -> return []
  is -> do
    let (iss, oss) = unzip $ map ports is
    iss' <- sequence [replicateM (length is) newPort | is <- iss]
    oss' <- sequence [replicateM (length os) newPort | os <- oss]
    emit =<< lift (component (concat $ iss ++ oss) (concat $ iss' ++ oss'))
    return [reconstruct i (is', os') | (i, is', os') <- zip3 is iss' oss']

type Cxt  = Var :-> Type
type Env  = Var :-> Interface
type LEnv = Var :-> Interface

contractInterfaces :: Interface -> (Interface, Interface) -> Compiler ()
contractInterfaces if0 (if1, if2) = do
  let (is0, os0) = ports if0
      (is1, os1) = ports if1
      (is2, os2) = ports if2
  emit =<< lift (contract (is0 ++ os0) (is1 ++ os1, is2 ++ os2))

contractEnvs :: Env -> Env -> Compiler Env
contractEnvs env1 env2 = do
  let f if1 if2 = do
        let (is1, os1) = ports if1
        is0 <- replicateM (length is1) newPort
        os0 <- replicateM (length os1) newPort
        let if0 = reconstruct if1 (is0, os0)
        contractInterfaces if0 (if1, if2)
        return if0
  contracted <- sequence
    [do x <- io; return (v, x)
    | (v, io) <- M.toList $ M.intersectionWith f env1 env2]
  let diff1 = M.difference env1 env2
      diff2 = M.difference env2 env1
  return $ M.fromList contracted `mappend` diff1 `mappend` diff2

compile :: TypedTerm ::: Type  -> Compiler (Interface, Env)
compile (term ::: typ) = case unTypedTerm term of
  Var v -> do
    interface           <- newInterface typ
    derelictedInterface <- addComponent derelict interface
    return (derelictedInterface, M.singleton v interface)
  Abs var (Just vtyp) m -> do
    (mInterface, env) <- compile m
    varInterface <- case M.lookup var env of
      -- If the variable is not used in the term, weaken it
      Nothing -> do
        interface <- newInterface vtyp
        let (pis, pos) = ports interface
        emit =<< lift (weaken $ pis ++ pos)
        return interface
      Just vi -> return vi
    return (In varInterface mInterface, M.delete var env)
  m :$ n -> do
    (mInterface, mEnv) <- compile m
    (nInterface, nEnv) <- compile n
    () <- trace (show (mInterface, nInterface)) $ return ()
    let (vs, inps)     = unzip $ M.toList nEnv
    (nInterface:inps)  <- addComponentIs box (nInterface : inps)
    inps               <- addComponentIs (flip comult) inps -- Why flip?
    cEnv               <- contractEnvs mEnv (M.fromList $ zip vs inps)
    mnInterface        <- apply mInterface nInterface
    return (mnInterface, cEnv)
  Lit n -> do
    interface@(Port MsgInt p) <- newInterface TInt
    emit =<< lift (int n p)
    return (interface, mempty)
  Plus -> do
    interface@(Port MsgInt p1d `In` (Port MsgInt p2d `In` Port MsgInt p3)) <-
      newInterface (TInt :->: TInt :->: TInt)
    p1 <- newPort; p2 <- newPort
    emit =<< lift (derelict [p1d] [p1])
    emit =<< lift (derelict [p2d] [p2])
    emit =<< lift (add (p1, p2) p3)
    return (interface, mempty)
  Minus -> do
    interface@(Port MsgInt p1d `In` (Port MsgInt p2d `In` Port MsgInt p3)) <-
      newInterface (TInt :->: TInt :->: TInt)
    p1 <- newPort; p2 <- newPort
    emit =<< lift (derelict [p1d] [p1])
    emit =<< lift (derelict [p2d] [p2])
    emit =<< lift (sub (p1, p2) p3)
    return (interface, mempty)
  Times -> do
    interface@(Port MsgInt p1d `In` (Port MsgInt p2d `In` Port MsgInt p3)) <-
      newInterface (TInt :->: TInt :->: TInt)
    p1 <- newPort; p2 <- newPort
    emit =<< lift (derelict [p1d] [p1])
    emit =<< lift (derelict [p2d] [p2])
    emit =<< lift (mul (p1, p2) p3)
    return (interface, mempty)
  Fix -> do
    interface@(In (In interface1 interface2) interfaceOut) <- newInterface typ
    interfaceOut' <- addComponent derelict interfaceOut
    interface1'   <- addComponent comult interface1
    contractInterfaces interface2 (interface1, interfaceOut)
    return (In (In interface1' interface2) interfaceOut', mempty)
  -- Case t brs def -> do
  If -> do
    interface@(Port MsgInt p1d `In` (Port MsgInt p2d `In` (Port MsgInt p3d `In` Port MsgInt p4))) <-
      newInterface (TInt :->: TInt :->: TInt :->: TInt)
    p1 <- newPort; p2 <- newPort; p3 <- newPort
    emit =<< lift (derelict [p1d] [p1])
    emit =<< lift (derelict [p2d] [p2])
    emit =<< lift (derelict [p3d] [p3])
    emit =<< lift (casee (p1, [(0, p3)], p2) p4)
    return (interface, mempty)
  m :@ node -> withNode node $ compile m

  where
    err = error $ "compile: " ++ show term
    elookup :: Ord k => k -> k :-> v -> v
    elookup v m = fromMaybe err $ M.lookup v m
