{-# LANGUAGE GeneralizedNewtypeDeriving, TypeOperators #-}
module Component where

import Control.Applicative
import Control.Arrow(second)
import Control.Monad.State
import Data.Char
import Data.Function
import Data.List
import Data.Map(Map)
import qualified Data.Map as M
import Data.Maybe
import Data.Monoid

infix  1 -->
infixr 5 :>
infixr 3 :->

type Port  = Int
type Node  = String
type Label = String
type a :-> b = Map a b

data ComponentCode
  = Listen (Port  :-> Label)
           (Label :-> Code)
  deriving (Eq, Ord, Show)

-- | Smart constructor for components
listen :: [(Port, Label)] -> [(Label, Code)] -> ComponentGen ComponentCode
listen ps ls = return $ Listen (M.fromList ps) (M.fromList ls)

data Code
  = Instr :> Code
  | Send Port
  | Jump Label
  | Match Label Label
  | Switch [(Int, Label)] Label
  | Done
  | Print
  deriving (Eq, Ord, Show)

data Instr
  = Inl   | Inr
  | Fst   | Snd
  | UnFst | UnSnd
  | Flip  | Push | Pop
  | Int Int | Add | Sub | Mul
  | PushInt
  deriving (Eq, Ord, Show)

data Data
  = DInl Data | DInr Data
  | Data :* Data
  | DEmpty
  | DInt Int
  deriving (Eq, Ord, Show)

prettyComponentCode :: ComponentCode -> String
prettyComponentCode (Listen ps ls) = "listen:\n"
  ++ intercalate "\n" [show p ++ " --> " ++ l | (p, l) <- M.toList ps] ++ "\n"
  ++ intercalate "\n" [l ++ ": " ++ prettyCode c | (l, c) <- M.toList ls]

prettyCode :: Code -> String
prettyCode code = case code of
  instr :> c  -> prettyInstr instr ++ "; " ++ prettyCode c
  Send p      -> "send " ++ show p
  Jump l      -> "jump " ++ l
  Match l1 l2 -> "match " ++ l1 ++ " " ++ l2
  Switch brs l -> "switch " ++ concat [show n ++ " -> " ++ l ++ "; " | (n, l) <- brs] ++ "_ -> " ++ l ++ ";"
  Done        -> "done"
  Print       -> "print"

prettyInstr :: Instr -> String
prettyInstr = map toLower . show

substPorts :: Port :-> Code -> Code -> Code
substPorts sub code = case code of
  instr :> c  -> instr :> substPorts sub c
  Send p      -> fromMaybe (Send p) $ M.lookup p sub
  Jump l      -> Jump l
  Match l1 l2 -> Match l1 l2
  Switch brs l -> Switch brs l
  Done        -> Done
  Print       -> Print

substPortsCC :: Port :-> Port -> ComponentCode -> ComponentCode
substPortsCC sub (Listen ps ls) =
  Listen (M.mapKeys pf ps)
         (M.map (substPorts $ M.map Send sub) ls)
  where pf p = fromMaybe p $ M.lookup p sub

(-->) :: a -> b -> (a, b)
x --> y = (x, y)

type ComponentState = [Integer]
initialComponentState :: ComponentState
initialComponentState = [0..]

newtype ComponentGen a = ComponentGen {unComponentGen :: State ComponentState a}
  deriving (Applicative, Functor, Monad, MonadState ComponentState)

newLabel :: ComponentGen String
newLabel = do
  i:is <- get
  put is
  return $ "l" ++ show i

labels :: [a] -> ComponentGen [Label]
labels ps = replicateM (length ps) newLabel

evalComponent :: ComponentGen a -> a
evalComponent = flip evalState initialComponentState . unComponentGen

type Component = ComponentGen ComponentCode

assertSameLength :: String -> [a] -> [a] -> ComponentGen ()
assertSameLength s a b = do
  when (length a /= length b) $ error $ s ++ ": lengths do not match"

ident :: [Port] -> [Port] -> Component
ident p0s p1s = do
  assertSameLength "ident" p0s p1s
  l0s <- labels p0s
  l1s <- labels p1s
  listen
    (zip (p0s ++ p1s) (l0s ++ l1s))
    ([ l0 --> Send p1 | (l0, p1) <- zip l0s p1s ] ++
     [ l1 --> Send p0 | (l1, p0) <- zip l1s p0s ])

derelict :: [Port] -> [Port] -> Component
derelict p0s p1s = do
  assertSameLength "derelict" p0s p1s
  l0s <- labels p0s
  l1s <- labels p1s
  listen
    (zip (p0s ++ p1s) (l0s ++ l1s))
    ([ l0 --> Snd :> Pop :> Send p1 | (l0, p1) <- zip l0s p1s ] ++
     [ l1 --> Push :> UnFst :> Send p0       | (l1, p0) <- zip l1s p0s ])

comult :: [Port] -> [Port] -> Component
comult p0s p1s = do
  assertSameLength "comult" p0s p1s
  l0s <- labels p0s
  l1s <- labels p1s
  listen
    (zip (p0s ++ p1s) (l0s ++ l1s))
    ([ l0 --> Fst :> Snd :> Flip :> UnFst :> UnSnd :> Send p1
     | (l0, p1) <- zip l0s p1s ] ++
     [ l1 --> Snd :> Fst :> Flip :> UnSnd :> UnFst :> Send p0
     | (l1, p0) <- zip l1s p0s ])

contract :: [Port] -> ([Port], [Port]) -> Component
contract p0s (p1s, p2s) = do
  assertSameLength "contract" p0s p1s
  assertSameLength "contract" p1s p2s
  l0s <- labels p0s
  l1s <- labels p1s
  l2s <- labels p2s
  lls <- labels p0s
  lrs <- labels p0s
  listen
    (zip (p0s ++ p1s ++ p2s) (l0s ++ l1s ++ l2s))
    (concat [
      [ l0 --> Fst   :> Match ll lr
      , ll --> UnFst :> Send p1
      , lr --> UnFst :> Send p2
      , l1 --> Fst   :> Inl :> UnFst :> Send p0
      , l2 --> Fst   :> Inr :> UnFst :> Send p0
      ] | (l0, l1, l2, ll, lr, p0, (p1, p2)) <- zip7 l0s l1s l2s lls lrs p0s (zip p1s p2s)])

weaken :: [Port] -> Component
weaken ps = do
  ls <- labels ps
  listen
    (zip ps ls)
    [l --> Done | l <- ls]


box :: [Port] -> [Port] -> Component
box inner outer = do
  assertSameLength "box" inner outer
  innerls <- labels inner
  outerls <- labels outer
  listen
    (zip inner innerls ++ zip outer outerls)
    $ concat [ [ol --> Snd :> Send ip, il --> UnSnd :> Send op]
             | (ip, op, il, ol) <- zip4 inner outer innerls outerls ]

add :: (Port, Port) -> Port -> Component
add (p0, p1) p2 = do
  [l0, l1, l2] <- labels [p0, p1, p2]
  listen
    (zip [p0, p1, p2] [l0, l1, l2])
    [ l2 --> Send p0
    , l0 --> PushInt :> Send p1
    , l1 --> Add :> Send p2
    ]

sub :: (Port, Port) -> Port -> Component
sub (p0, p1) p2 = do
  [l0, l1, l2] <- labels [p0, p1, p2]
  listen
    (zip [p0, p1, p2] [l0, l1, l2])
    [ l2 --> Send p0
    , l0 --> PushInt :> Send p1
    , l1 --> Sub :> Send p2
    ]

mul :: (Port, Port) -> Port -> Component
mul (p0, p1) p2 = do
  [l0, l1, l2] <- labels [p0, p1, p2]
  listen
    (zip [p0, p1, p2] [l0, l1, l2])
    [ l2 --> Send p0
    , l0 --> PushInt :> Send p1
    , l1 --> Mul :> Send p2
    ]

casee :: (Port, [(Int, Port)], Port) -> Port -> Component
casee (scrutinee, brs, def) out = do
  [lscrutinee,ldefin,lin] <- labels [scrutinee, def, out]
  lbrsin    <- labels brs
  lbrsout   <- labels brs
  [ldefout] <- labels [def]
  listen
    (zip (scrutinee:def:out:map snd brs) (lscrutinee:ldefin:lin:lbrsin))
    $ concat [ [l   --> Send out | l <- ldefin:lbrsin]
             , [lin --> Send scrutinee]
             , [lscrutinee --> Switch (zip (map fst brs) lbrsout) ldefout]
             , [l   --> Send p | (l, p) <- zip (ldefout : lbrsout) (def : map snd brs)]
             ]

int :: Int -> Port -> Component
int n p = do
  l <- newLabel
  listen
    [(p, l)]
    [ l --> Int n :> Send p ]

print :: Port -> Component
print p0 = do
  l0 <- newLabel
  listen
    [(p0, l0)]
    [ l0 --> Print ]

combine :: ComponentCode -> ComponentCode -> ComponentCode
combine (Listen ps1 ls1) (Listen ps2 ls2) =
  if M.null (M.intersection ls1' ls2') then Listen (ps1' `mappend` ps2')
                                                   (ls1' `mappend` ls2')
    else error "combine: ls1' intersects ls2'"
  where
    ips1 = M.intersection ps1 ps2
    ips2 = M.intersection ps2 ps1
    ps1' = M.filterWithKey (\k _ -> M.notMember k ips1) ps1
    ps2' = M.filterWithKey (\k _ -> M.notMember k ips2) ps2
    ls1' = M.map (substPorts $ M.map Jump ips2) ls1
    ls2' = M.map (substPorts $ M.map Jump ips1) ls2

test = evalComponent $ do
  c1 <- contract [0] ([1], [2])
  c2 <- contract [2] ([3], [4])
  return $ prettyComponentCode $ combines [c1, c2]

test2 = evalComponent $ do
  box <- box [0,1,2] [3,4,5]
  c1  <- contract [0] ([1], [2])
  return $ prettyComponentCode $ combines [c1, box]

combines :: [ComponentCode] -> ComponentCode
combines = foldr1 combine

combineComponents :: [(ComponentCode, Maybe Node)] -> [(ComponentCode, Node)]
combineComponents ccs = combined ++ unspecified
  where (mspecified, munspecified) = partition (isJust . snd) ccs
        grouped                    = groupBy ((==) `on` snd)
                                   $ sortBy (compare `on` snd)
                                   $ map (second fromJust) mspecified
        combined                   = map go grouped
        go ccns@((cc, n):_)        = (combines $ map fst ccns, n)
        unspecified                = zip (map fst munspecified)
                                   $ map show [0..]
